#include <string.h>
#include <math.h>
#include <mex.h>
#include <assert.h>

void minass_climb(double *rq, double *x, double *H, double *c, double *A, int ntoskip, int maxsteps, int n, int m, int nm,int verbosity);
double linass(double *x, int *ass, double *c, int *start, int *stop, int ntoskip, int n, int m, int nm, int verbosity);
double grad_descent(double *x, int *ass, double *H, double *c, int *start, int *stop, int ntoskip, int n, int m, int nm, int nsteps, int verbosity);

void mexFunction (int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[]) {

  // check input arguments, use them to call the real routine

  if ( ( nrhs !=6 ) || ( nlhs != 2 ) ) {
    mexErrMsgTxt("usage [x,q] = minass_climb(H,c,spA,ntoskip,maxsteps,verbosity)\n'");
  }

  // set pointers to the arguments

  double* H = mxGetPr(prhs[0]);
  double* c = mxGetPr(prhs[1]);
  double* A = mxGetPr(prhs[2]);
  int ntoskip = (int)mxGetScalar(prhs[3]);
  int maxsteps = (int)mxGetScalar(prhs[4]);
  int verbosity = (int)mxGetScalar(prhs[5]);
  int n = mxGetM(prhs[0]);
  int m = mxGetM(prhs[2]);


  // check argument sizes
  
  if ( mxGetM(prhs[0]) != mxGetM(prhs[0]) ) { 
    mexErrMsgTxt("H should be square.\n"); 
  }
  if ( mxGetM(prhs[0]) != mxGetM(prhs[1]) ) { 
    mexErrMsgTxt("c should be a column vector with the same dimension as H .\n"); 
  }
  //  if ( mxGetM(prhs[0]) != mxGetN(prhs[2]) ) {
  //      mexErrMsgTxt("A should have the same number of columns as H .\n");
  //  }
  //  if ( mxGetM(prhs[2]) != mxGetM(prhs[3]) ) {    
  //    mexErrMsgTxt("b should have the same number of rows as A .\n");
  //  }

  
  if (verbosity > 0){
    fprintf(stdout,"size of H = %dx%d  rows in Aeq = %d  maxsteps = %d verbosity = %d\n",n,n,m,maxsteps,verbosity);
  }



  int nm = m;  // convenience variable, number of model points == number of rows of Aeq except the last one
                 // that provides the constraint for how many model points are assigned to null


  // initialize output argument

  plhs[0] = mxCreateDoubleMatrix(n,1,mxREAL);
  plhs[1] = mxCreateDoubleMatrix(n,1,mxREAL);

  double* x = mxGetPr(plhs[0]);
  double* rq = mxGetPr(plhs[1]);


  // Call the real routine

  minass_climb(rq,x,H,c,A,ntoskip,maxsteps,n,m,nm,verbosity);


  fprintf(stdout,"rq[0] = %4.3f\n",rq[0]);
}


void minass_climb(double *rq, double *x, double *H, double *c, double *A, int ntoskip, int maxsteps, int n, int m, int nm, int verbosity){
  if (verbosity > 1) {
    fprintf(stdout, "In minass_climb\n");
  }

  // extract the starts and stop points

  int *start = new int[nm];
  assert(start);
  int *stop  = new int[nm];
  assert(stop);

  for (int i=0; i<nm; i++){
    start[i] = (int) (A[i]);
    stop[i] = (int)  (A[i+nm]);
    
    if (verbosity>2){
      fprintf(stdout, "start[%d] = %d\n",i,start[i]);
    }

    if (verbosity>2){
      fprintf(stdout, "stop[%d] = %d\n",i,stop[i]);
    }
  }

  if (verbosity>1){
    fprintf(stdout, "ntoskip = %d\n",ntoskip);
  }
  // Consider each possible assignment


  int *ass = new int[nm];

  double *indep_costs = new double[n];
  assert(indep_costs);
  for (int i=0; i<n; i++){    
  //  for (int i=0; i<1; i++){    
    indep_costs[i] = linass(x,ass,&(H[i*n]),start,stop,ntoskip,n,m,nm,verbosity) + c[i];
    if (verbosity>2){
      fprintf(stdout,"indep_costs[%d] = %4.2f.\n",i,indep_costs[i]);
    }
  }

  fprintf(stdout,"n was %d.\n",n);
  fprintf(stdout,"n was %d.\n",n);
  for (int i=0; i<n; i++){
    rq[i]=indep_costs[i];
    //    if (verbosity>2){
    //      fprintf(stdout,"rq[%d] = %4.2f.\n",i,rq[i]);
    //    }
  }


  double min_lin = linass(x,ass,indep_costs,start,stop,ntoskip,n,m,nm,verbosity);
  fprintf(stdout,"min_lin = %4.2f\n",min_lin);
  

  double min_quad = grad_descent(x,ass,H,c,start,stop,ntoskip,n,m,nm,maxsteps,verbosity);

  fprintf(stdout,"min_quad = %4.2f\n",min_quad);

  delete ass;
  delete indep_costs;
  delete start;
  delete stop;
  
}



double grad_descent(double *x, int *ass, double *H, double *c, int *start, int *stop, int ntoskip, int n, int m, int nm, int nsteps, int verbosity){
  if (verbosity > 2) {
    fprintf(stdout, "In grad_descent\n");
  }

    if (verbosity > 3) {
      for (int i=0; i<nm; i++){
	fprintf(stderr, "%d %d\n",i,ass[i]);
      }
    }

    double *old_x = new double[n];
    for (int i=0; i<n; i++){
      old_x[i] = x[i];
    }
			   
			     
  int *nb = new int[nm];
  for (int i=0; i<nm; i++){
    nb[i] = ass[i];
  }    
  double nb_score = 0;
  for (int i=0; i<nm; i++){
    double *tp = &(H[nb[i]*n]);
    for (int j=0; j<nm; j++){
      nb_score = nb_score + tp[nb[j]];
    }
    nb_score = nb_score + c[nb[i]];
  }



  int *curr = new int[nm];
  double curr_score = 0;

  int *next = new int[nm];
  int *temp_ass = new int[nm];
  int *temp_temp_ass = new int[nm];
  double next_score = 0;

  fprintf(stderr,"Hello bef nsteps=%d.\n",nsteps);
  for (int step=0; step<nsteps; step++){



    for (int i=0; i<nm; i++){
      curr[i] = nb[i];
    }
    curr_score = nb_score;


    if (verbosity > -1) {
      fprintf(stderr, "\nstep %d current cost %4.2f\n",step,curr_score);
    }

    for (int j=0; j<nm; j++){
      next[j] = curr[j];
    }    
    next_score = curr_score;
  

    for (int i=0; i<nm; i++){                     // which  feature
      for (int lcv=0; lcv<nm; lcv++){
	temp_ass[lcv]=curr[lcv];
      }
      for (int j=start[i]; j<=stop[i]; j++){       // which assignment

	// Check the singleton change... if it isn't to null
	temp_ass[i]=j;


	  double pot_score = 0;

	  /*
	  for (int k=0; k<nm; k++){
	    double *tp = &(H[temp_ass[k]*n]);
	    for (int l=0; l<nm; l++){
	      pot_score = pot_score + tp[temp_ass[l]];
	    }
	    pot_score = pot_score + c[temp_ass[k]];
	  }
	  */


	  double score_change = 0;
	  for (int k=0; k<nm; k++){
            // add the new cost
	    score_change = score_change + 2*H[nb[k]*n+temp_ass[i]];
            // subtract the old cost
	    score_change = score_change - 2*H[nb[k]*n+nb[i]];
            // take care of the overlaps
	  }
	  score_change = score_change + H[nb[i]*n+nb[i]];
	  score_change = score_change + H[temp_ass[i]*n+temp_ass[i]];
	  score_change = score_change - 2*H[temp_ass[i]*n+nb[i]];
	  // add on the change in linear term
	  score_change = score_change + c[temp_ass[i]];
	  score_change = score_change - c[nb[i]];


	  //	  fprintf(stdout,"just for fun %4.2f and %4.2f.\n",pot_score,curr_score+score_change);

	  pot_score = curr_score+score_change;

	  double outer_pot_score = pot_score;

	if (j<stop[i]){	  
	  if (pot_score<next_score){
	    next_score = pot_score;
	    for (int k=0; k<nm; k++){
	      next[k] = temp_ass[k];
	    }     
	    fprintf(stderr,".");
	  }
	}

       	// Check for a double change...
	for (int i2=i; i2<nm; i2++){                      // which second
	  for (int k=0; k<nm; k++){
	    temp_temp_ass[k] = temp_ass[k];
	  }
	  for (int j2=start[i2]; j2<=stop[i2]; j2++){       // which assignment for second feature
	    temp_temp_ass[i2]=j2;
	    if ( ((j<stop[i])&&(j2<stop[i2])) || ((j2==stop[i2])&&(j<stop[i])&&(curr[i]==stop[i])) \
		 ||   ((j==stop[i])&&(j2<stop[i2])&&(curr[i2]==stop[i2]))   ){
	      double pot_score = 0;
	      /*
	      for (int k=0; k<nm; k++){
		double *tp = &(H[temp_temp_ass[k]*n]);
		for (int l=0; l<nm; l++){
		  pot_score = pot_score + tp[temp_temp_ass[l]];
		}
		pot_score = pot_score + c[temp_temp_ass[k]];
	      }
	      */

	      double score_change = 0;
	      for (int k=0; k<nm; k++){
		// add the new cost
		score_change = score_change + 2*H[temp_ass[k]*n+temp_temp_ass[i2]];
		// subtract the old cost
		score_change = score_change - 2*H[temp_ass[k]*n+temp_ass[i2]];
		// take care of the overlaps
	      }
	      score_change = score_change + H[temp_ass[i2]*n+temp_ass[i2]];
	      score_change = score_change + H[temp_temp_ass[i2]*n+temp_temp_ass[i2]];
	      score_change = score_change - 2*H[temp_temp_ass[i2]*n+temp_ass[i2]];
	      // add on the change in linear term
	      score_change = score_change + c[temp_temp_ass[i2]];
	      score_change = score_change - c[temp_ass[i2]];
	      

	      //	      	  fprintf(stdout,"just for fun %4.2f and %4.2f.\n",pot_score,outer_pot_score+score_change);
	      
	      pot_score = outer_pot_score+score_change;


	      if (pot_score<next_score){
		next_score = pot_score;
		for (int k=0; k<nm; k++){
		  next[k] = temp_temp_ass[k];
		}     
		fprintf(stderr,".");
	      }
	    }
	  }
	}
      }
    }
    if (curr_score==next_score){
      fprintf(stderr,"No more changes.\n");
      break;
    }
    int some_change=0;
    for (int i=0; i<nm; i++){
      if (nb[i]!=next[i]){
	some_change=1;
      }
      nb[i] = next[i];
    }
    if (!some_change){
      fprintf(stdout,"No more changes.\n");
      break;
    }
    nb_score = next_score;
  }
  fprintf(stdout,"skipping bad.\n");

  fprintf(stdout,"skipping good.\n");
  for (int i=0; i<n; i++){
    x[i]=0;
  }

  for (int i=0; i<nm; i++){
    x[nb[i]]=1;
  }

  for (int i=0; i<n; i++){
    if (x[i]!=old_x[i]){
      fprintf(stdout,"Changed x[%d] from %2.0f to %2.0f.\n",i,old_x[i],x[i]);
    }
  }


  double r=next_score;

  delete old_x;
  delete nb;
  delete curr;
  delete next;
  delete temp_ass;
  delete temp_temp_ass;
  fprintf(stderr,"\n");

  return(r);

}


double linass(double *x, int *ass, double *c, int *start, int *stop, int ntoskip, int n, int m, int nm, int verbosity){
  if (verbosity > 2) {
    fprintf(stdout, "In linass\n");
  }

  double *null_benefit = new double[nm];
  assert(null_benefit);
  double *cost = new double[nm];
  assert(cost);
  int *min_inds = new int[nm];
  assert(min_inds);
  int *inds = new int[ntoskip];
  assert(inds);
  double *sorted_benefit = new double[ntoskip];
  assert(sorted_benefit);

  //  for (int j=0; j<stop[nm-1]; j++){
  //    fprintf(stderr,"%4.2f ",c[j]);
  //  }
  //  fprintf(stderr,"\n");

  for (int i=0; i<nm; i++){
    int my_start = start[i];
    //    fprintf(stderr,"start[%d]=%d.\n",i,start[i]);
    int my_stop = stop[i];
    //    fprintf(stderr,"stop[%d]=%d.\n",i,stop[i]);
    double min_val = c[my_start];
    int    min_ind = my_start;
    for (int j=(my_start+1); j<my_stop; j++){
      double val = c[j];
      if ( val < min_val ){
	min_val = val;
	min_ind = j;
      }
    }
    cost[i] = min_val;
    //    fprintf(stderr,"cost[%d]=%4.2f.\n",i,cost[i]);
    min_inds[i] = min_ind;
    null_benefit[i] = min_val-c[my_stop];
  }

  
  int nsorted = 0;
  for (int i=0; i<ntoskip; i++){
    if (i==nm){
      break;
    }
    sorted_benefit[i] = null_benefit[i];
    inds[i]=i;
    nsorted++;
  }
  
  
  double min_val = 100000;
  int min_val_ind = -1;
  if (nsorted>0){
    min_val = sorted_benefit[0];
    min_val_ind = 0;
  }
  for (int i=1; i<nsorted; i++){
    double val = sorted_benefit[i];
    if ( val < min_val ){
      min_val = val;
      min_val_ind = i;
    }
  }

  for (int i=ntoskip; i<nm; i++){
    double val = null_benefit[i];
    if (val > min_val){
      sorted_benefit[min_val_ind]=val;
      inds[min_val_ind]=i;
      min_val = sorted_benefit[0];
      min_val_ind = 0;
      for (int j=1; j<nsorted; j++){
	double jval = sorted_benefit[j];
	if ( jval < min_val ){
	  min_val = jval;
	  min_val_ind = j;
	}
      }
    }
  }

  for (int i=0; i<n; i++){
    x[i]=0;
  }
  
  double r=0;
  for (int i=0; i<nm; i++){
    r = r + cost[i];
    x[min_inds[i]]=1;
    ass[i] = min_inds[i];
  }

  //  fprintf(stderr,"r = %4.2f\n",r);

  for (int i=0; i<nsorted; i++){
    r = r - sorted_benefit[i];
    //    fprintf(stderr,"sorted_benefit[%d]=%4.3f\n",i,sorted_benefit[i]);
    x[min_inds[inds[i]]]=0;
    x[stop[inds[i]]]=1;
    ass[inds[i]] = stop[inds[i]];
  }

  //  fprintf(stderr,"r = %4.2f\n",r);

  delete sorted_benefit;
  delete inds;
  delete cost;
  delete min_inds;
  delete null_benefit;
  return(r);
}

    
//     // setup local variables





//     int *minp = new int[nm];
//     int *maxp = new int[nm];

    
//     // find the range of possibile assignments for each model point

//     for (int i=0; i<nm; i++){
//       int j;
//       for (j=0; j<n; j++){
// 	if (A[i+j*m]>0.999){
// 	  minp[i] = j;
// 	  to1[i] = (double)(minp[i]);
// 	  //	  printf("%d ",minp[i]);
// 	  break;
// 	}
//       }

//       for ( j++; j<n; j++){
// 	if (A[i+j*m]<0.999){
// 	  maxp[i] = j-1;
// 	  //	  printf("%d \n",maxp[i]);
// 	  	  to2[i] = (double)(maxp[i]);
// 	  break;
// 	}
//       }
//       //      printf("%d.\n",j);
//       if (j==n){
// 	maxp[i]=j-1;
// 	// printf("%d \n",maxp[i]);
//       	to2[i] = (double)(maxp[i]);
//       }
//     }


//     // solve for the q_i s


//     // count and find the nulls

//     int *nullsv = new int[n];
//     int nnulls = 0;
//     for (int i=0; i<n; i++){
//       nullsv[i] = (int)A[i*m+m-1];
//       if (nullsv[i]){
// 	nnulls = nnulls+1;
//       }
//     }
    
//     int *nulls = new int[nnulls];
//     int nullcount=0;
//     for (int i=0; i<n; i++){
//       if  ((int)A[i*m+m-1]){
// 	nulls[nullcount]=i;
// 	nullcount = nullcount+1;

// 	//	printf("%d\n",i);
//       }
//     }
    

//     int ntoskip = (int) b[m-1];
//     printf("ntoskip = %d\n",ntoskip);
//     int nonulls = (ntoskip==0);


//     //    double *u = new double[n];
//     for (int i=0; i<n; i++){
//       u[i]=0;
//       x[i]=0;
//       x1[i]=0;
//     }

//     printf("\n\n");
//     double *col = new double[n];
//     for (int i=0; i<n; i++){
//       if ((nonulls)&&(nullsv[i])){
// 	u[i] = 1000000; // a big discouraging number
//       }
//       else{
// 	for (int j=0; j<n; j++){ // copy out a col of H so that it is conveniently accessible....???
// 	  col[j] = H[i*n+j];
// 	}
// 	for (int j=0; j<n; j++){x[j]=0;}
// 	 u[i] = linass(ntoskip,  nm,  n, m, col, minp, maxp, x);
// 	 //	u[i]=1;
// 	if  (i==0){
// 	  printf("Hello Here.\n");
// 	  for (int k=0; k<n; k++){
// 	    x1[k]=x[k];
// 	  }
// 	}
// 	//	for (int j=0; j<n; j++){
// 	//	  printf("x_%d[%d] = %4.3f\n",i,j,x[j]);
// 	//	}
// 	//	printf("%4.2f\n",u[i]);
//       }
//     }
//     //    for (int i=0; i<n; i++){
//     //      printf("u[%d] = %4.3f\n",i,u[i]);
//     //    }
//     for (int i=0; i<n; i++){
//       u[i] = u[i] + c[i];
//     }

//     linass(ntoskip,  nm,  n, m, u, minp, maxp, x);


//     int maxxs = 1000043;
//     int nsx = 0;
//     int *xs = new int[nm*maxxs];
//     for (int i=0; i<(nm*maxxs); i++){
//       xs[i]=0;
//     }

//     int *curr_x = new int[nm];
//     int max_hashes = 100000007;
//     unsigned char *hashes = new unsigned char[max_hashes+1];
//     long int hash_count=0;
//     double *QP = new double[maxxs];
//     double *LP = new double[maxxs];
//     for (int i=0; i<(maxxs); i++){
//       QP[i]=0;
//       LP[i]=0;
//     }

    
//     for (int i=0; i<(max_hashes+1); i++){
//       hashes[i]=0;
//     }

//     double qps = 0;
//     double lps = 0;
//     for (int i=0; i<nm; i++){
//       for (int j=minp[i]; j<=maxp[i]; j++){
// 	if (x[j]){
// 	  xs[i]=j;
// 	  lps = lps + u[j];
// 	  qps = qps + c[j];
// 	  break;
// 	}
//       }
//     }
//     for (int i=0; i<nm; i++){
//       for (int j=0; j<nm; j++){
// 	qps = qps + H[xs[i]*n+xs[j]];
//       }
//     }

//     LP[0]= lps;
//     QP[0]= qps;
//     hashes[hashit(xs,nm)]=1;
//     hash_count = hash_count+1;
//     nsx = nsx+1;
    
//     double curr_lp = 0;
//     double max_lp = LP[0];
//     double curr_qp = 0;
//     double best_qp = QP[0]; 
//     double gap=QP[0]-LP[0];
//     long int curr_hash=-1;
//     int *best_x = new int[nm];
//     for (int i=0; i<nm; i++){
//       best_x[i] = xs[i];
//     }
//     int done=0;
//     int remove_count =0;
//     while (!done){
//       double mv = LP[0];
//       int mi = 0;
//       for (int i=1; i<nsx; i++){
// 	if (LP[i]<mv){
// 	  mv = LP[i];
// 	  mi = i;
// 	}
//       }
//       printf("Removing %d  nsx = %d.\n",mi,nsx);

//       for (int i=0; i<nm; i++){
// 	curr_x[i] = xs[mi*nm+i];
//       }
//       curr_lp = LP[mi];
//       max_lp = curr_lp;
//       curr_qp = QP[mi];
//       if (mi<(nsx-1)){
// 	LP[mi]=LP[nsx-1];
// 	QP[mi]=QP[nsx-1];
// 	for (int i=0; i<nm; i++){
// 	  xs[mi*nm+i]=xs[(nsx-1)*nm+i];
// 	}
//       }
//       nsx = nsx-1;
//       remove_count = remove_count+1;

//       gap = best_qp-curr_lp;
//       printf("Gap was %4.2f \t QP was %4.2f \t LP was %4.10f.\n",gap,best_qp,curr_lp);

//       if (gap<=0){
// 	printf("Found minima.\n");
//       }
//       if ((gap<=0)||(remove_count>maxstorage)){
// 	for (int i=0; i<n; i++){
// 	  x[i]=0;
// 	}
// 	for (int i=0; i<nm; i++){
// 	  x[best_x[i]]=1;
// 	}
// 	goto done_label;
//       }
    
//       // add points around curr_x

//       for (int i=0; i<nm; i++){
//         int skip_sum=0;
// 	for (int cc=0; cc<nm; cc++){
// 	  if (cc!=i){
// 	    skip_sum = skip_sum + (curr_x[cc]==maxp[cc]);
// 	  }
// 	}
	
// 	for (int np=minp[i]; np < (maxp[i]-(skip_sum>=ntoskip)); np++){
// 	  if (np!=curr_x[i]){
// 	    if ((nsx>=(maxxs-1))){
// 	      done=1;

// 	      for (int i=0; i<n; i++){
// 		x[i]=0;
// 	      }
// 	      for (int i=0; i<nm; i++){
// 		x[best_x[i]]=1;
// 	      }
	      
// 	      goto done_label;

// 	    }
// 	    for (int j=0; j<nm; j++){
// 	      xs[nsx*nm+j]=curr_x[j];
// 	    }
// 	    xs[nsx*nm+i] = np;
	    
// 	    // check to see if this has already been stored...
// 	    int chash = hashit(&(xs[nsx*nm]),nm);
// 	    int found = (hashes[chash]);
	    
// 	    hashes[chash]=1;

// 	    nsx = nsx + 1;
// 	    hash_count++;

// 	    if (found){
// 	      nsx = nsx-1;
// 	      hash_count = hash_count-1;
// 	    }else{ 
// 	      QP[nsx-1]=0;
// 	      LP[nsx-1]=0;
// 	      for (int j=0; j<nm; j++){
// 		for (int k=0; k<nm; k++){
// 		  QP[nsx-1]=QP[nsx-1]+H[xs[(nsx-1)*nm+j]*n+xs[(nsx-1)*nm+k]];
// 		}
// 		QP[nsx-1]=QP[nsx-1]+c[xs[(nsx-1)*nm+j]];
// 		LP[nsx-1]=LP[nsx-1]+u[xs[(nsx-1)*nm+j]];
// 	      }
// 	      if (QP[nsx-1]<best_qp){
// 		best_qp = QP[nsx-1];
// 		for (int sss=0; sss<nm; sss++){
// 		  best_x[sss]=xs[(nsx-1)*nm+sss];
// 		}
// 		gap = best_qp-max_lp;
// 	      }
// 	    }
// 	  }
// 	}
//       }
//     }
    
//     for (int i=0; i<n; i++){
//       x[i]=0;
//     }
//     for (int i=0; i<nm; i++){
//       x[best_x[i]]=1;
//     }
    

    
//  done_label:

//     delete xs;
//     delete curr_x;
//     delete hashes;
//     delete QP;
//     delete LP;
//     delete best_x;

//  next_to_last_label:


//     delete minp;
//     delete maxp;
//     delete col;
//     delete nullsv;
//     delete nulls;

// }


// double linass(int ntoskip,  int nm,  int n, int  m, double *col, int *minp, int *maxp, double *x){
//   double *sorted_vals = new double[ntoskip];
//   int *sorted_inds = new int[ntoskip];
//   double *skip=new double[nm];
//   double *v=new double[nm];
//   int *vi=new int[nm];
//   double *va=new double[nm];
//   int *vai=new int[nm];

//   double u=0;

//   for ( int j=0; j<n ;j++){
//     x[j]=0;
//   }

//   for ( int j=0; j<nm; j++){
//     int start = minp[j];
//     int stop = maxp[j]-1;
//     int mi=start;
//     double mv = col[start];
//     for ( int k=start+1; k<=stop; k++){  // find the min in this interval...
//       if (col[k]<mv){
// 	mv = col[k];
// 	mi = k;
//       }
//     }
//     va[j] = mv;
//     vai[j] = mi;

//     if (col[maxp[j]]<mv){
//       mv = col[maxp[j]];
//       mi = maxp[j];
//     }
//     skip[j] = va[j]-mv;

//     v[j] = mv;
//     vi[j] = mi;

//   }

//   if (ntoskip>0){
//     int nin = 1;  
//     sorted_vals[0]=skip[0];
//     sorted_inds[0]=0;
//     u=0;
//     for (int j=1; j<nm; j++){
//       if ((nin < ntoskip)||(skip[j]>sorted_vals[nin-1])){
// 	int k = 0;
// 	for (k=0; k<nin; k++){
// 	  if (skip[j]>sorted_vals[k]){
// 	    break;
// 	  }
// 	}
// 	// printf("k was %d\n",k);
//        	if (k>=ntoskip){
//      	  printf("Should not be greater than ntoskip...\n");
//  	}
// 	double temp = sorted_vals[k];
// 	int tempi = sorted_inds[k];
// 	sorted_vals[k]=skip[j];
// 	sorted_inds[k]=j;
// 	if (nin<ntoskip){
// 	  nin = nin+1;
// 	}
// 	int l;
// 	for (l=k+1; (l<ntoskip)&&(l<nin); l++){
// 	  double temp2 = sorted_vals[l];
// 	  int tempi2 = sorted_inds[l];
// 	  sorted_vals[l]=temp;
// 	  sorted_inds[l]=tempi;
// 	  temp = temp2;
// 	  tempi = tempi2;
// 	}
//       }
//     }
// //    for (int j=0; j<ntoskip; j++){
// //      printf("%4.2f\n",sorted_vals[j]);
// //    }
//     for (int j=0; j<ntoskip; j++){
//       va[sorted_inds[j]] = v[sorted_inds[j]];
//       //      if (va[sorted_inds[j]]<=0){
//       //	printf("Hello, funny %4.2f.\n", v[sorted_inds[j]]);
//       //      }
//       vai[sorted_inds[j]] = vi[sorted_inds[j]];
//     }
//   }

//   for (int j=0; j<nm; j++){
//     x[vai[j]]=1;
//     //    printf("setting x[%d]\n",vai[j]);
//     u = u + va[j];
//   }


//   delete sorted_vals;
//   delete sorted_inds;
//   delete skip;
//   delete v;
//   delete vi;
//   delete va;
//   delete vai;

//   return(u);
// }


// long int hashit( int *xs, int nm){
//   int max_hash = 100000007;
//   int mult1 = 63689;
//   int mult2 = 378551;

//   long int hv = 0;
//   long int mult = mult1;
//   for (int j=0; j<nm; j++){
//     hv = hv*mult+ xs[j];
//     hv = (long int)((double)hv - (double)(int)(hv/max_hash)*(double)max_hash);
//     hv = hv + 1;
//     mult = mult*mult2;
//     mult = mult - (int)(mult/max_hash)*max_hash;
//   }
//   return(hv);
// }

