% compute shapeme vocabulary by sampling pb edge points with f/g labels
% train figure/ground classifier using learned shapemes
function fg_classifier = shapeme_vocab(fg_data, type)
   % default type - training data
   if (nargin < 2), type = 'train'; end
   % set gb parameters
   rs      = [4 6 8 11];
   nthetas = [16 16 16 16];
   alpha   = 0.5;
   beta_gb = 1;
   % repulsion radius for gb
   rrep = 3;
   % maximum number of descriptors per image
   n_samp = 1000;
   % sigma for classifier distance function
   sigma = 0.15;
   % set directories
   pb_dir = ['../data/' type '/pb_components'];
   fg_dir = ['../data/' type '/fg_pb_corr'];
   % get number of images
   n_images = numel(fg_data);
   % initialize descriptor storage
   gbs         = zeros([(n_images.*n_samp) sum(nthetas)]); % descriptors
   gbs_label   = zeros([(n_images.*n_samp) 1]); % f/g label
   gbs_ori_ind = zeros([(n_images.*n_samp) 1]); % original orientation index
   gbs_inds    = zeros([(n_images.*n_samp) 1]);
   n_desc_im = zeros([n_images 1]);
   n_desc = 0;
   % compute descriptors for each image
   for n = 1:n_images
      % read image
      iid = fg_data(n).iid;
      im = imgRead(iid,'color');
      % read pb components
      pb = load([pb_dir '/' num2str(iid) '.mat']);
      pb = mPb_all_from_cues( ...
         pb.bg1, pb.bg2, pb.bg3, ...
         pb.cga1, pb.cga2, pb.cga3, ...
         pb.cgb1, pb.cgb2, pb.cgb3, ...
         pb.tg1, pb.tg2, pb.tg3, 1.0);
      [pbm ori] = max(pb,[],3);
      % read f/g correspondance
      FG = load([fg_dir '/' num2str(iid) '_fg_pb.mat']);
      FG = FG.FG;
      [sx sy] = size(FG.edge_fg);
      % sample descriptor locations
      [ci cj] = find_interest_points(double(FG.edge_fg ~= 0), n_samp, rrep);
      n_curr = numel(ci);
      inds = sub2ind([sx sy],ci,cj);
      % compute gb descriptors
      [descriptors sample_points] = ...
         compute_gb(FG.nmax,ci,cj,alpha,beta_gb,rs,nthetas);
      % normalize descriptors
      descriptors = normalize_descriptors(descriptors,rs,nthetas);
      % rotate descriptors
      drot = zeros(size(descriptors));
      for idx = 1:n_curr
         frot = reshape(descriptors(idx,:),[16 4]);
         frot = circshift(frot,[(ori(inds(idx))-1) 0]);
         drot(idx,:) = reshape(frot,[1 64]);
      end
      % store descriptors
      gbs((n_desc+1):(n_desc+n_curr),:)       = drot;
      gbs_label((n_desc+1):(n_desc+n_curr))   = FG.edge_fg(inds);
      gbs_ori_ind((n_desc+1):(n_desc+n_curr)) = ori(inds);
      % store indices
      gbs_inds((n_desc+1):(n_desc+n_curr)) = inds;
      % interactively select and show descriptors
      if (false)
         figure(99); imagesc(FG.edge_fg); axis image;
         figure(10); imagesc(im); axis image;
         figure(11); imagesc(pbm); axis image;
         hold on; plot(cj,ci,'r*'); hold off;
         while (true)
            % select descriptor
            figure(11);
            [x y b] = ginput_range([sx sy]);
            if (b == 2), break; end
            % display descriptor at selected point
            dist = (x-ci).*(x-ci) + (y-cj).*(y-cj);
            [dmin idx] = min(dist);
            figure(11); imagesc(pbm); axis image;
            hold on; plot(cj,ci,'r*'); hold off;
            hold on; plot(cj(idx),ci(idx),'w*'); hold off;
            hold on; plot(cj(idx)+sample_points(:,2),ci(idx)+sample_points(:,1),'wo'); hold off;
            feature = descriptors(idx,:,:);
            % draw descriptor
            figure(12);
            draw_descriptor(squeeze(feature),sample_points);
            % draw rotated descriptor
            frot = reshape(feature,[16 4]);
            frot = circshift(frot,[(ori(inds(idx))-1) 0]);
            disp(['rotation = ' num2str((ori(inds(idx))-1))]);
            figure(13);
            draw_descriptor(reshape(frot,[64 1]),sample_points);
         end
      end
      % increment count
      n_desc_im(n) = n_curr;
      n_desc = n_desc + n_curr;
   end
   % trim unusued storage
   gbs         = gbs(1:n_desc,:);
   gbs_label   = gbs_label(1:n_desc);
   gbs_ori_ind = gbs_ori_ind(1:n_desc);
   gbs_inds    = gbs_inds(1:n_desc);
   % cluster gb descriptors
   k = 64;
   [IDX C SUMD D] = kmeans(gbs,k);
   % count frequency of figure label in each cluster
   ffreq = zeros([k 1]);
   for n = 1:k
      mems = find(IDX==n);
      ffreq(n) = numel(find(gbs_label(mems)==1))./numel(mems);
   end
   % show cluster centers
   for n = 1:k
      figure(1); subplot(8,8,n);
      draw_descriptor(C(n,:),sample_points);
      title([num2str(ffreq(n)) '    ' num2str(numel(find(IDX==n)))]);
   end
   % train a logistic classifier on square distances
   F = [exp(-D./sigma) ones([n_desc 1])];
   label = gbs_label.*(gbs_label>0);
   [beta p lli] = logist2(label,F);
   % return shapemes and classifier parameters
   fg_classifier.shape_centroids = C;
   fg_classifier.beta            = beta;
   fg_classifier.sigma           = sigma;
   % record training accuracy
   fg_classifier.train_acc = numel(find(round(p)==label))./numel(p);

   if (false)
      % show predicted f/g labels
      pos = 0;
      for n = 1:n_images
         % read image
         iid = fg_data(n).iid;
         im = imgRead(iid,'color');
         % read pb components 
         pb = load([pb_dir '/' num2str(iid) '.mat']);
         pb = mPb_all_from_cues( ...
            pb.bg1, pb.bg2, pb.bg3, ...
            pb.cga1, pb.cga2, pb.cga3, ...
            pb.cgb1, pb.cgb2, pb.cgb3, ...
            pb.tg1, pb.tg2, pb.tg3, 1.0);
         [pbm ori] = max(pb,[],3);
         % read f/g correspondance
         FG = load([fg_dir '/' num2str(iid) '_fg_pb.mat']);
         FG = FG.FG;
         [sx sy] = size(FG.edge_fg);
         % get indices and predictions
         n_curr = n_desc_im(n);
         inds  = gbs_inds((pos+1):(pos+n_curr));
         preds = p((pos+1):(pos+n_curr));
         % plot f/g predictions
         [xcs ycs] = ind2sub([sx sy],inds);
         os = (FG.oris(FG.ori_ind(inds)) + pi/2).';
         os = os + pi.*(preds < 0.5);
         xcs_a = xcs + (FG.radius).*abs(preds-0.5).*8.*cos(os);
         ycs_a = ycs + (FG.radius).*abs(preds-0.5).*8.*sin(os);
         figure(3); imagesc(FG.nmax); axis image; colormap(gray);
         hold on;
         plot([ycs ycs_a].',[xcs xcs_a].','g');
         plot([ycs_a].',[xcs_a].','w.');
         hold off;
         % update position
         pos = pos + n_curr;
         % compute connection locations
         xa = round(xcs + 2.*(FG.radius).*cos(os));
         ya = round(ycs + 2.*(FG.radius).*sin(os));
         xb = round(xcs - 2.*(FG.radius).*cos(os));
         yb = round(ycs - 2.*(FG.radius).*sin(os));
         g = find( ...
            (xa > 1) & (xa < sx) & (xb > 1) & (xb < sx) & ...
            (ya > 1) & (ya < sy) & (yb > 1) & (yb < sy));
         ia = sub2ind([sx sy],xa(g),ya(g));
         ib = sub2ind([sx sy],xb(g),yb(g));
         vals = repmat(pi./8,[numel(ia) 1]);
         conf = 2.*abs(preds(g)-0.5).*FG.nmax(inds(g));
         index_a = cat(1, ia, ib);
         index_b = cat(1, ib, ia);
         vals_all = cat(1, -vals, vals);
         conf_all = cat(1, conf, conf);
         rho = 0.1;
         A = pb_to_affinity(pb, rho);
         % create sparse ordering matrices
         R = sparse(index_a, index_b, vals_all, (sx*sy), (sx*sy));
         C = sparse(index_a, index_b, conf_all, (sx*sy), (sx*sy));
         %% compute eigenvectors for simulataneous grouping and f/g problem
         nvec = 16;
         [evecs_ae evals_ae] = ncut_ae(A + 0.0001.*C, R, nvec);
         %% compute weights
         weights_ae = zeros(size(evals_ae));
         for v = 1:nvec
            if (abs(evals_ae(v)) > 0)
               weights_ae(v) = 1./sqrt(abs(evals_ae(v)));
            end
         end
         weights_ae = weights_ae./sum(weights_ae);
         %% sort by weight
         [junk inds] = sort(weights_ae);
         weights_ae = weights_ae(inds(end:-1:1));
         evecs_ae   = evecs_ae(:,inds(end:-1:1));
         %% display f/g
         figure(1); imagesc(im); axis image;
         figure(2); imagesc(reshape(angle(evecs_ae(:,1)),[sx sy])); axis image; title('ae vect 1 (angle)'); colorbar;
         pause;
      end
   end
end
