% test local shapeme-based figure/ground classifier accuracy
function acc = shapeme_vocab_test(fg_data, fg_classifier, type)
   % default type - training data
   if (nargin < 3), type = 'train'; end
   % set gb parameters
   rs      = [4 6 8 11];
   nthetas = [16 16 16 16];
   alpha   = 0.5;
   beta_gb = 1;
   % repulsion radius for gb
   rrep = 3;
   % maximum number of descriptors per image
   n_samp = 1000;
   % sigma for classifier distance function
   sigma = 0.15;
   % set directories
   pb_dir = ['../data/' type '/pb_components'];
   fg_dir = ['../data/' type '/fg_pb_corr'];
   % get number of images
   n_images = numel(fg_data);
   % initialize descriptor storage
   gbs         = zeros([(n_images.*n_samp) sum(nthetas)]); % descriptors
   gbs_label   = zeros([(n_images.*n_samp) 1]); % f/g label
   gbs_ori_ind = zeros([(n_images.*n_samp) 1]); % original orientation index
   gbs_inds    = zeros([(n_images.*n_samp) 1]);
   n_desc_im = zeros([n_images 1]);
   n_desc = 0;
   % compute descriptors for each image
   for n = 1:n_images
      % read image
      iid = fg_data(n).iid;
      im = imgRead(iid,'color');
      % read pb components
      pb = load([pb_dir '/' num2str(iid) '.mat']);
      pb = mPb_all_from_cues( ...
         pb.bg1, pb.bg2, pb.bg3, ...
         pb.cga1, pb.cga2, pb.cga3, ...
         pb.cgb1, pb.cgb2, pb.cgb3, ...
         pb.tg1, pb.tg2, pb.tg3, 1.0);
      [pbm ori] = max(pb,[],3);
      % read f/g correspondance
      FG = load([fg_dir '/' num2str(iid) '_fg_pb.mat']);
      FG = FG.FG;
      [sx sy] = size(FG.edge_fg);
      % sample descriptor locations
      [ci cj] = find_interest_points(double(FG.edge_fg ~= 0), n_samp, rrep);
      n_curr = numel(ci);
      inds = sub2ind([sx sy],ci,cj);
      % compute gb descriptors
      [descriptors sample_points] = ...
         compute_gb(FG.nmax,ci,cj,alpha,beta_gb,rs,nthetas);
      % normalize descriptors
      descriptors = normalize_descriptors(descriptors,rs,nthetas);
      % rotate descriptors
      drot = zeros(size(descriptors));
      for idx = 1:n_curr
         frot = reshape(descriptors(idx,:),[16 4]);
         frot = circshift(frot,[(ori(inds(idx))-1) 0]);
         drot(idx,:) = reshape(frot,[1 64]);
      end
      % store descriptors
      gbs((n_desc+1):(n_desc+n_curr),:)       = drot;
      gbs_label((n_desc+1):(n_desc+n_curr))   = FG.edge_fg(inds);
      gbs_ori_ind((n_desc+1):(n_desc+n_curr)) = ori(inds);
      % store indices
      gbs_inds((n_desc+1):(n_desc+n_curr)) = inds;
      % interactively select and show descriptors
      if (false)
         figure(99); imagesc(FG.edge_fg); axis image;
         figure(10); imagesc(im); axis image;
         figure(11); imagesc(pbm); axis image;
         hold on; plot(cj,ci,'r*'); hold off;
         while (true)
            % select descriptor
            figure(11);
            [x y b] = ginput_range([sx sy]);
            if (b == 2), break; end
            % display descriptor at selected point
            dist = (x-ci).*(x-ci) + (y-cj).*(y-cj);
            [dmin idx] = min(dist);
            figure(11); imagesc(pbm); axis image;
            hold on; plot(cj,ci,'r*'); hold off;
            hold on; plot(cj(idx),ci(idx),'w*'); hold off;
            hold on; plot(cj(idx)+sample_points(:,2),ci(idx)+sample_points(:,1),'wo'); hold off;
            feature = descriptors(idx,:,:);
            % draw descriptor
            figure(12);
            draw_descriptor(squeeze(feature),sample_points);
            % draw rotated descriptor
            frot = reshape(feature,[16 4]);
            frot = circshift(frot,[(ori(inds(idx))-1) 0]);
            disp(['rotation = ' num2str((ori(inds(idx))-1))]);
            figure(13);
            draw_descriptor(reshape(frot,[64 1]),sample_points);
         end
      end
      % increment count
      n_desc_im(n) = n_curr;
      n_desc = n_desc + n_curr;
   end
   % trim unusued storage
   gbs         = gbs(1:n_desc,:);
   gbs_label   = gbs_label(1:n_desc);
   gbs_ori_ind = gbs_ori_ind(1:n_desc);
   gbs_inds    = gbs_inds(1:n_desc);
   % compute gb to shapeme distances
   k = size(fg_classifier.shape_centroids,1);
   dmx = zeros([n_desc k]);
   for idx = 1:n_desc
      dist = repmat(gbs(idx,:),[k 1]) - fg_classifier.shape_centroids;
      dist = sum(dist.*dist,2);
      dmx(idx,:) = dist;
   end
   % form feature matrix
   F = [exp(-dmx./(fg_classifier.sigma)) ones([n_desc 1])];
   % predict f/g
   preds = 1./(1+exp(-F*(fg_classifier.beta)));
   % compute accuracy
   label = gbs_label.*(gbs_label>0);
   acc = numel(find(round(preds)==label))./numel(preds);
end
